"""Implementation the TELBO algorithm from "Generative Models of Visually Grounded Imagination"
    (https://arxiv.org/abs/1705.10762)
"""

from .telbo_config import TELBOConfig
from .telbo_model import TELBO
